import os
import re
import subprocess
from dataclasses import dataclass

import modal
from loguru import logger
from modal import method

from sweepai.utils.config.server import UTILS_MODAL_INST_NAME

stub = modal.Stub(UTILS_MODAL_INST_NAME)
tiktoken_image = modal.Image.debian_slim().pip_install("tiktoken", "loguru", "anthropic", "pyyaml", "PyGithub")

TIKTOKEN_CACHE_DIR = "/root/cache/tiktoken"
tiktoken_volume = modal.SharedVolume().persist("tiktoken-models")


@stub.cls(
    image=tiktoken_image,
    shared_volumes={TIKTOKEN_CACHE_DIR: tiktoken_volume},
    secret=modal.Secret.from_dict({"TIKTOKEN_CACHE_DIR": TIKTOKEN_CACHE_DIR})
)
class Tiktoken:
    openai_models = ["gpt-3.5-turbo", "gpt-4", "gpt-4-32k", "gpt-4-32k-0613"]
    anthropic_models = ["claude-v1", "claude-v1.3-100k", "claude-instant-v1.3-100k"]
    models = openai_models + anthropic_models

    def __enter__(self):
        import tiktoken
        self.openai_models = {model: tiktoken.encoding_for_model(model) for model in Tiktoken.openai_models}

    @method()
    def count(self, text: str, model: str = "gpt-4"):
        return len(self.openai_models[model].encode(text))


chunking_image = modal.Image.debian_slim() \
    .apt_install("git") \
    .pip_install("tree-sitter", "loguru", "pyyaml", "PyGithub")

CHUNKING_CACHE_DIR = "/root/cache/"
chunking_volume = modal.SharedVolume().persist("chunking-parsers")


@dataclass
class Span:
    start: int
    end: int

    def extract(self, s: str) -> str:
        return "\n".join(s.splitlines()[self.start:self.end])

    def __add__(self, other):
        if isinstance(other, int):
            return Span(self.start + other, self.end + other)
        elif isinstance(other, Span):
            return Span(self.start, other.end)
        else:
            raise NotImplementedError()

    def __len__(self):
        return self.end - self.start


def get_line_number(index: int, source_code: str) -> int:
    # unoptimized, use binary search
    lines = source_code.splitlines(keepends=True)
    total_chars = 0
    line_number = 0
    while total_chars <= index:
        if line_number == len(lines):
            return line_number
        total_chars += len(lines[line_number])
        line_number += 1
    return line_number - 1


def chunker(tree, source_code_bytes, max_chunk_size=512 * 3, coalesce=50):
    # Recursively form chunks with a maximum chunk size of max_chunk_size
    def chunker_helper(node, source_code_bytes, start_position=0):
        chunks = []
        current_chunk = Span(start_position, start_position)
        for child in node.children:
            child_span = Span(child.start_byte, child.end_byte)
            if len(child_span) > max_chunk_size:
                chunks.append(current_chunk)
                chunks.extend(chunker_helper(child, source_code_bytes, child.start_byte))
                current_chunk = Span(child.end_byte, child.end_byte)
            elif len(current_chunk) + len(child_span) > max_chunk_size:
                chunks.append(current_chunk)
                current_chunk = child_span
            else:
                current_chunk += child_span
        if len(current_chunk) > 0:
            chunks.append(current_chunk)
        return chunks

    chunks = chunker_helper(tree.root_node, source_code_bytes)

    # removing gaps
    for prev, curr in zip(chunks[:-1], chunks[1:]):
        prev.end = curr.start

    # combining small chunks with bigger ones
    new_chunks = []
    i = 0
    current_chunk = Span(0, 0)
    while i < len(chunks):
        current_chunk += chunks[i]
        if count_length_without_whitespace(
                source_code_bytes[current_chunk.start:current_chunk.end].decode("utf-8")) > coalesce \
                and "\n" in source_code_bytes[current_chunk.start:current_chunk.end].decode("utf-8"):
            new_chunks.append(current_chunk)
            current_chunk = Span(chunks[i].end, chunks[i].end)
        i += 1
    if len(current_chunk) > 0:
        new_chunks.append(current_chunk)

    line_chunks = [Span(get_line_number(chunk.start, source_code=source_code_bytes),
                        get_line_number(chunk.end, source_code=source_code_bytes)) for chunk in new_chunks]
    line_chunks = [chunk for chunk in line_chunks if len(chunk) > 0]

    return line_chunks


def count_length_without_whitespace(s: str):
    string_without_whitespace = re.sub(r'\s', '', s)
    return len(string_without_whitespace)


extension_to_language = {
    "js": "tsx",
    "jsx": "tsx",
    "ts": "tsx",
    "tsx": "tsx",
    "mjs": "tsx",
    "py": "python",
    "rs": "rust",
    "go": "go",
    "java": "java",
    "cpp": "cpp",
    "cc": "cpp",
    "cxx": "cpp",
    "c": "cpp",
    "h": "cpp",
    "hpp": "cpp",
    "cs": "c-sharp",
    "rb": "ruby",
    "md": "markdown",
    "rst": "markdown",
    "txt": "markdown",
    "erb": "embedded-template",
    "ejs": "embedded-template",
    "html": "embedded-template",
    "vue": "vue",
    "php": "php",
}


@stub.cls(
    image=chunking_image,
    shared_volumes={CHUNKING_CACHE_DIR: chunking_volume},
)
class Chunking:

    def __enter__(self):
        from tree_sitter import Language

        logger.debug("Downloading tree-sitter parsers")

        LANGUAGE_NAMES = ["python", "java", "cpp", "go", "rust", "ruby", "php"]
        for language in LANGUAGE_NAMES:
            subprocess.run(
                f"git clone https://github.com/tree-sitter/tree-sitter-{language} cache/tree-sitter-{language}",
                shell=True)
        for language in LANGUAGE_NAMES:
            Language.build_library(f'cache/build/{language}.so', [f"cache/tree-sitter-{language}"])
            subprocess.run(f"cp cache/build/{language}.so /tmp/{language}.so", shell=True)  # copying for executability
        self.languages = {language: Language(f"/tmp/{language}.so", language) for language in LANGUAGE_NAMES}

        subprocess.run(f"git clone https://github.com/tree-sitter/tree-sitter-typescript cache/tree-sitter-typescript",
                       shell=True)
        Language.build_library(f'cache/build/typescript.so', [f"cache/tree-sitter-typescript/tsx"])
        subprocess.run(f"cp cache/build/typescript.so /tmp/typescript.so", shell=True)
        self.languages["tsx"] = Language("/tmp/typescript.so", "tsx")

        subprocess.run(f"git clone https://github.com/tree-sitter/tree-sitter-c-sharp cache/tree-sitter-c-sharp",
                       shell=True)
        Language.build_library(f'cache/build/c-sharp.so', [f"cache/tree-sitter-c-sharp"])
        subprocess.run(f"cp cache/build/c-sharp.so /tmp/c-sharp.so", shell=True)
        self.languages["c-sharp"] = Language("/tmp/c-sharp.so", "c_sharp")

        subprocess.run(
            f"git clone https://github.com/tree-sitter/tree-sitter-embedded-template cache/tree-sitter-embedded-template",
            shell=True)
        Language.build_library(f'cache/build/embedded-template.so', [f"cache/tree-sitter-embedded-template"])
        subprocess.run(f"cp cache/build/embedded-template.so /tmp/embedded-template.so", shell=True)
        self.languages["embedded-template"] = Language("/tmp/embedded-template.so", "embedded_template")

        subprocess.run(f"git clone https://github.com/MDeiml/tree-sitter-markdown cache/tree-sitter-markdown",
                       shell=True)
        Language.build_library(f'cache/build/markdown.so', [f"cache/tree-sitter-markdown/tree-sitter-markdown"])
        subprocess.run(f"cp cache/build/markdown.so /tmp/markdown.so", shell=True)
        self.languages["markdown"] = Language("/tmp/markdown.so", "markdown")

        subprocess.run(f"git clone https://github.com/ikatyang/tree-sitter-vue cache/tree-sitter-vue", shell=True)
        Language.build_library(f'cache/build/vue.so', [f"cache/tree-sitter-vue"])
        subprocess.run(f"cp cache/build/vue.so /tmp/vue.so", shell=True)
        self.languages["vue"] = Language("/tmp/vue.so", "vue")

        logger.debug("Finished downloading tree-sitter parsers")

    @method()
    def chunk(
            self,
            file_content: str,
            file_path: str,
            score: float = 1.0,
            additional_metadata: dict[str, str] = {},
            max_chunk_size: int = 512 * 3,
            chunk_size: int = 30,
            overlap: int = 15
    ) -> tuple[list[str], list[dict[str, str]]]:
        """This function returns a list of chunks and a list of metadata for each chunk.
        chunks: list of file chunks
        metadata: list of metadata for each chunk example: {"file_path": "python", "start": 0, "end": 10}
        ids: list of ids for each chunk {file_path}:{start}{end}
        """
        # TODO(Sweep): implement a config file for the above
        # TODO(Sweep): Prioritize the language based on extension
        from tree_sitter import Parser
        file_language = None
        tree = None

        logger.info(f"Chunking {file_path}, size {len(file_content)}")

        _, ext = os.path.splitext(file_path)
        ext = ext[len("."):]
        if ext in extension_to_language:
            # prioritize the language
            language_names = [extension_to_language[ext]]
            language_names += [
                language_name for language_name in self.languages.keys()
                if language_name != extension_to_language[ext]
            ]
            logger.info(language_names)
        else:
            language_names = list(self.languages.keys())

        for language_name in language_names:
            language = self.languages[language_name]
            parser = Parser()
            parser.set_language(language)
            tree = parser.parse(bytes(file_content, "utf-8"))
            if not tree.root_node.children or tree.root_node.children[0].type != "ERROR":
                file_language = language
                break
            logger.warning(f"Not language {language_name}")

        ids = []
        metadatas = []

        if file_language:
            logger.info(file_language.name)
            source_code_bytes = bytes(file_content, "utf-8")
            spans = chunker(tree, source_code_bytes, max_chunk_size)
            ids = [f"{file_path}:{span.start}:{span.end}" for span in spans]
            chunks = [span.extract(file_content) for span in spans]
            for chunk in chunks:
                print(chunk + "\n\n\n")
            for span in spans:
                metadata = {
                    "file_path": file_path,
                    "start": span.start,
                    "end": span.end,
                    "score": score,
                    **additional_metadata
                }
                metadatas.append(metadata)
        else:
            # start and end refers to line here, will fix later
            logger.info("Unknown language")
            source_lines = file_content.split('\n')
            num_lines = len(source_lines)
            logger.info(f"Number of lines: {num_lines}")
            chunks = []
            start_line = 0
            while start_line < num_lines and num_lines > overlap:
                end_line = min(start_line + chunk_size, num_lines)
                chunk = '\n'.join(source_lines[start_line:end_line])
                chunks.append(chunk)
                ids.append(f"{file_path}:{start_line}:{end_line}")
                metadatas.append({
                    "file_path": file_path,
                    "start": start_line,
                    "end": end_line,
                    "score": score,
                    **additional_metadata
                })
                start_line += chunk_size - overlap

        return chunks, metadatas, ids



import json
import webbrowser
from typing import Any

import httpx
import requests
from loguru import logger
from pydantic import BaseModel

from sweepai.app.config import SweepChatConfig
from sweepai.core.entities import Function, PullRequest, Snippet
from sweepai.utils.config.client import GITHUB_APP_CLIENT_ID, SWEEP_API_ENDPOINT

create_pr_function = Function(
    name="create_pr",
    description="Creates a PR.",
    parameters={
        "properties": {
            "plan": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "file_path": {
                            "type": "string",
                            "description": "The file path to change."
                        },
                        "instructions": {
                            "type": "string",
                            "description": "Concise NATURAL LANGUAGE summary of what to change in each file. There should be absolutely NO code, only English.",
                            "example": [
                                "Refactor the algorithm by moving the main function to the top of the file.",
                                "Change the implementation to recursion"
                            ]
                        },
                    },
                    "required": ["file_path", "instructions"]
                },
                "description": "A list of files to modify or create and corresponding instructions."
            },
            "title": {
                "type": "string",
                "description": "Title of PR",
            },
            "summary": {
                "type": "string",
                "description": "Detailed summary of PR",
            },
            "branch": {
                "type": "string",
                "description": "Name of branch to create PR in.",
            },
        },
        "required": ["plan", "title", "summary", "branch"]
    }
)

create_pr_function_call = {"name": "create_pr"}


def break_json(raw_json: str):
    # turns something like {"function_call": {"arguments": " \""}}{"function_call": {"arguments": "summary"}} into two objects
    try:
        yield json.loads(raw_json)
    except json.JSONDecodeError:
        for i in range(1, len(raw_json)):
            try:
                obj = json.loads(raw_json[:i])
                yield obj
                for item in break_json(raw_json[i:]):
                    yield item
                break
            except json.JSONDecodeError:
                pass


class APIClient(BaseModel):
    config: SweepChatConfig
    api_endpoint = SWEEP_API_ENDPOINT

    def __init__(self, config: SweepChatConfig):
        super().__init__(config=config)
        self.config = config
        logger.info(f"Initializing API client")
        logger.info(f"API endpoint: {self.api_endpoint}")
        logger.info(f"Github APP Client ID: {GITHUB_APP_CLIENT_ID}")

    def get_installation_id(self):
        results = requests.post(
            self.api_endpoint + "/installation_id",
            json=self.config.dict(),
        )
        if results.status_code == 401:
            print("Installation ID not found! Please install sweep first.")
            webbrowser.open_new_tab("https://github.com/apps/sweep-ai")
            raise Exception(results.json()["detail"])
        if results.status_code != 200:
            raise Exception(results.json()["detail"])
        obj = results.json()
        return obj["installation_id"]

    def search(
            self,
            query: str,
            n_results: int = 5,
    ):
        results = requests.post(
            self.api_endpoint + "/search",
            json={
                "query": query,
                "n_results": n_results,
                "config": self.config.dict(),
            }
        )
        if results.status_code != 200:
            raise Exception(results.text)
        snippets = [Snippet(**item) for item in results.json()]
        return snippets

    def create_pr(
            self,
            file_change_requests: list[tuple[str, str]],
            pull_request: PullRequest,
            messages: list[tuple[str | None, str | None]],
    ):
        results = requests.post(
            self.api_endpoint + "/create_pr",
            json={
                "file_change_requests": file_change_requests,
                "pull_request": pull_request,
                "messages": messages,
                "config": self.config.dict(),
            },
            timeout=10 * 60
        )
        return results.json()

    def chat(
            self,
            messages: list[tuple[str | None, str | None]],
            snippets: list[Snippet] = [],
            model: str = "gpt-4-0613",
    ) -> str:
        results = requests.post(
            self.api_endpoint + "/chat",
            json={
                "messages": messages,
                "snippets": [snippet.dict() for snippet in snippets],
                "config": self.config.dict()
            }
        )
        return results.json()

    def stream_chat(
            self,
            messages: list[tuple[str | None, str | None]],
            snippets: list[Snippet] = [],
            functions: list[Function] = [],
            function_call: Any = "auto",
            model: str = "gpt-4-0613"
    ):
        with httpx.Client(timeout=30) as client:  # sometimes this step is slow
            with client.stream(
                    'POST',
                    self.api_endpoint + '/chat_stream',
                    json={
                        "messages": messages,
                        "snippets": [snippet.dict() for snippet in snippets],
                        "do_add_plan": True,
                        "functions": [func.dict() for func in functions],
                        "function_call": function_call,
                        "config": self.config.dict()
                    }
            ) as response:
                for delta_chunk in response.iter_text():
                    if not delta_chunk:
                        break
                    try:
                        for item in break_json(delta_chunk):
                            yield item
                    except json.decoder.JSONDecodeError as e:
                        logger.error(delta_chunk)
                        raise e



import modal
from loguru import logger

from sweepai.core.entities import Snippet
from sweepai.utils.config.server import UTILS_MODAL_INST_NAME

chunker = modal.Function.lookup(UTILS_MODAL_INST_NAME, "Chunking.chunk")


def format_snippets(snippets: list[Snippet]):
    snippets: list[Snippet] = snippets[::-1]

    num_full_files = 3
    num_extended_snippets = 5

    most_relevant_snippets = snippets[-num_full_files:]
    snippets = snippets[:-num_full_files]
    logger.info("Expanding snippets...")
    for snippet in most_relevant_snippets:
        current_snippet = snippet
        _chunks, metadatas, _ids = chunker.call(
            current_snippet.content,
            current_snippet.file_path
        )
        segmented_snippets = [
            Snippet(
                content=current_snippet.content,
                start=metadata["start"],
                end=metadata["end"],
                file_path=metadata["file_path"],
            ) for metadata in metadatas
        ]
        index = 0
        while index < len(segmented_snippets) and segmented_snippets[index].start <= current_snippet.start:
            index += 1
        index -= 1
        for i in range(index + 1, min(index + num_extended_snippets + 1, len(segmented_snippets))):
            current_snippet += segmented_snippets[i]
        for i in range(index - 1, max(index - num_extended_snippets - 1, 0), -1):
            current_snippet = segmented_snippets[i] + current_snippet
        snippets.append(current_snippet)
    return snippets



from datetime import datetime


def compute_score(contents, commits):
    line_count = contents.count("\n")
    if line_count > 200:
        line_count_score = 10
    else:
        line_count_score = line_count / 20
    commit_count = len(commits) + 1
    days_since_last_modified = max(((datetime.now() - commits[0].commit.author.date).total_seconds() // 3600), 0) + 1
    return (line_count_score * commit_count / days_since_last_modified)


def convert_to_percentiles(values):
    sorted_values = sorted(values)  # Sort the values in ascending order
    n = len(sorted_values)
    max_percentile = .25
    percentile_mapping = {value: (i / (n)) * max_percentile for i, value in enumerate(sorted_values)}
    percentiles = [percentile_mapping[value] for value in values]  # Create the percentiles list based on the mapping

    return percentiles



import os
import re
import string
from typing import ClassVar, Literal, Type, TypeVar

from loguru import logger
from pydantic import BaseModel

Self = TypeVar("Self", bound="RegexMatchableBaseModel")


class Message(BaseModel):
    role: Literal["system"] | Literal["user"] | Literal["assistant"] | Literal["function"]
    content: str | None = None
    name: str | None = None
    function_call: dict | None = None
    key: str | None = None

    @classmethod
    def from_tuple(cls, tup: tuple[str | None, str | None]) -> Self:
        if tup[0] is None:
            return cls(role="assistant", content=tup[1])
        else:
            return cls(role="user", content=tup[0])

    def to_openai(self) -> str:
        obj = {
            "role": self.role,
            "content": self.content,
        }
        if self.function_call:
            obj["function_call"] = self.function_call
        if self.role == "function":
            obj["name"] = self.name
        return obj


class Function(BaseModel):
    class Parameters(BaseModel):
        type: str = "object"
        properties: dict

    name: str
    description: str
    parameters: Parameters


class RegexMatchError(ValueError):
    pass


class RegexMatchableBaseModel(BaseModel):
    _regex: ClassVar[str]

    @classmethod
    def from_string(cls: Type[Self], string: str, **kwargs) -> Self:
        # match = re.search(file_regex, string, re.DOTALL)
        match = re.search(cls._regex, string, re.DOTALL)
        if match is None:
            logger.warning(f"Did not match {string} with pattern {cls._regex}")
            raise RegexMatchError("Did not match")
        return cls(
            **{k: (v if v else "").strip() for k, v in match.groupdict().items()},
            **kwargs,
        )


class FilesToChange(RegexMatchableBaseModel):
    files_to_modify: str
    files_to_create: str
    _regex = r"""<create>(?P<files_to_create>.*)</create>\s*<modify>(?P<files_to_modify>.*)</modify>"""


# todo (fix double colon regex): Update the split from "file_tree.py : desc" to "file_tree.py\tdesc"
# tab supremacy
def clean_filename(file_name: str):
    valid_chars = "-_./[]%s%s" % (string.ascii_letters, string.digits)
    file_name = ''.join(c for c in file_name if c in valid_chars)
    file_name = file_name.replace(' ', '')
    return os.path.normpath(file_name)


def clean_instructions(instructions: str):
    return instructions.strip()


class FileChangeRequest(RegexMatchableBaseModel):
    filename: str
    instructions: str
    change_type: Literal["modify"] | Literal["create"]

    @classmethod
    def from_string(cls: Type[Self], string: str, **kwargs) -> Self:
        colon_idx = string.find(':')
        file_name = string[:colon_idx]
        instructions = string[colon_idx + 1:]
        file_name = clean_filename(file_name)
        instructions = clean_instructions(instructions)
        res = FileChangeRequest(filename=file_name,
                                instructions=instructions,
                                change_type="modify")
        return res


class FileCreation(RegexMatchableBaseModel):
    commit_message: str
    code: str
    _regex = r'''commit_message\s+=\s+"(?P<commit_message>.*?)".*?<new_file>(python|javascript|typescript|csharp|tsx|jsx)?(?P<code>.*)<\/new_file>'''

    # _regex = r"""Commit Message:(?P<commit_message>.*)<new_file>(python|javascript|typescript|csharp|tsx|jsx)?(?P<code>.*)$"""
    # _regex = r"""Commit Message:(?P<commit_message>.*)(<new_file>|```)(python|javascript|typescript|csharp|tsx|jsx)?(?P<code>.*)($|```)"""

    @classmethod
    def from_string(cls: Type[Self], string: str, **kwargs) -> Self:
        result = super().from_string(string, **kwargs)
        result.code = result.code.strip()
        if result.code.endswith("</new_file>"):
            result.code = result.code[: -len("</new_file>")]
        if len(result.code) == 1:
            result.code = result.code.replace("```", "")
            return result.code + "\n"
        if result.code.startswith("```"):
            first_newline = result.code.find("\n")
            last_newline = result.code.rfind("\n")
            result.code = result.code[first_newline + 1:]
            result.code = result.code[: last_newline]
        result.code += "\n"
        return result


class PullRequest(RegexMatchableBaseModel):
    title: str
    branch_name: str
    content: str
    _regex = r'''pr_title\s+=\s+"(?P<title>.*?)"\n+branch\s+=\s+"(?P<branch_name>.*?)"\n+pr_content\s+=\s+"""(?P<content>.*?)"""'''


class Snippet(BaseModel):
    """
    Start and end refer to line numbers
    """

    content: str
    start: int
    end: int
    file_path: str

    def get_snippet(self):
        snippet = "\n".join(self.content.splitlines()[self.start:self.end])
        if self.start > 1:
            snippet = '...\n' + snippet
        if self.end < self.content.count('\n') + 1:
            snippet = snippet + '\n...'
        return snippet

    def __add__(self, other):
        assert self.content == other.content
        assert self.file_path == other.file_path
        return Snippet(
            content=self.content,
            start=self.start,
            end=other.end,
            file_path=self.file_path
        )

    def __xor__(self, other: "Snippet") -> bool:
        """
        Returns True if there is an overlap between two snippets.
        """
        if self.file_path != other.file_path:
            return False
        return self.file_path == other.file_path and (
                (self.start <= other.start and self.end >= other.start)
                or (other.start <= self.start and other.end >= self.start)
        )

    def __or__(self, other: "Snippet") -> "Snippet":
        assert self.file_path == other.file_path
        return Snippet(
            content=self.content,
            start=min(self.start, other.start),
            end=max(self.end, other.end),
            file_path=self.file_path
        )

    @property
    def xml(self):
        return f"""<snippet filepath="{self.file_path}" start="{self.start}" end="{self.end}">\n{self.get_snippet()}\n</snippet>"""

    def get_url(self, repo_name: str, commit_id: str = "main"):
        num_lines = self.content.count("\n") + 1
        return f"https://github.com/{repo_name}/blob/{commit_id}/{self.file_path}#L{max(self.start, 1)}-L{min(self.end, num_lines)}"

    def get_markdown_link(self, repo_name: str, commit_id: str = "main"):
        num_lines = self.content.count("\n") + 1
        base = commit_id + "/" if commit_id != "main" else ""
        return f"[{base}{self.file_path}#L{max(self.start, 1)}-L{min(self.end, num_lines)}]({self.get_url(repo_name, commit_id)})"

    def get_slack_link(self, repo_name: str, commit_id: str = "main"):
        num_lines = self.content.count("\n") + 1
        base = commit_id + "/" if commit_id != "main" else ""
        return f"<{self.get_url(repo_name, commit_id)}|{base}{self.file_path}#L{max(self.start, 1)}-L{min(self.end, num_lines)}>"

    def get_preview(self, max_lines: int = 5):
        snippet = "\n".join(self.content.splitlines()[self.start:min(self.start + max_lines, self.end)])
        if self.start > 1:
            snippet = '\n' + snippet
        if self.end < self.content.count('\n') + 1 and self.end > max_lines:
            snippet = snippet + '\n'
        return snippet

    def expand(self, num_lines: int = 50):
        return Snippet(
            content=self.content,
            start=max(self.start - num_lines, 1),
            end=min(self.end + num_lines, self.content.count("\n") + 1),
            file_path=self.file_path
        )

    @property
    def denotation(self):
        return f"{self.file_path}:{self.start}-{self.end}"


class DiffSummarization(RegexMatchableBaseModel):
    content: str
    _regex = r"""<file_summarization>(?P<content>.*)<\/file_summarization>"""


class PullRequestComment(RegexMatchableBaseModel):
    changes_required: str
    content: str
    _regex = r"""<changes_required>(?P<changes_required>.*)<\/changes_required>(\s+)<review_comment>(?P<content>.*)<\/review_comment>"""


class NoFilesException(Exception):
    def __init__(self, message="Sweep could not find any files to modify"):
        super().__init__(message)



import json
from datetime import datetime, timedelta
from typing import Any

from fastapi import requests
from loguru import logger
from pydantic import BaseModel, Field
from pymongo import MongoClient

from sweepai.utils.config.server import MONGODB_URI, DISCORD_WEBHOOK_URL


class ChatLogger(BaseModel):
    data: dict = Field(default_factory=dict)
    chat_collection: Any = None
    ticket_collection: Any = None
    expiration: datetime = None
    index: int = 0
    current_month: str = datetime.utcnow().strftime('%m/%Y')

    def __init__(self, data: dict = Field(default_factory=dict)):
        super().__init__(data=data)  # Call the BaseModel's __init__ method
        key = MONGODB_URI
        if key is None:
            logger.warning('Chat history logger has no key')
            return
        try:
            client = MongoClient(key, serverSelectionTimeoutMS=5000, socketTimeoutMS=5000)
            db = client['llm']
            self.chat_collection = db['chat_history']
            self.ticket_collection = db['tickets']
            self.ticket_collection.create_index('username')
            self.chat_collection.create_index('expireAt', expireAfterSeconds=0)
            self.expiration = datetime.utcnow() + timedelta(days=1)
        except Exception as e:
            logger.warning('Chat history could not connect to MongoDB')
            logger.warning(e)

    def get_chat_history(self, filters):
        return self.chat_collection.find(filters) \
            .sort([('expiration', 1), ('index', 1)]) \
            .limit(2000)

    def add_chat(self, additional_data):
        document = {**self.data, **additional_data, 'expiration': self.expiration, 'index': self.index}
        self.index += 1
        self.chat_collection.insert_one(document)

    def add_successful_ticket(self):
        if self.ticket_collection is None:
            logger.error('Ticket Collection Does Not Exist')
            return
        username = self.data['username']
        self.ticket_collection.update_one(
            {'username': username},
            {'$inc': {self.current_month: 1}},
            upsert=True
        )
        logger.info(f'Added Successful Ticket for {username}')

    def get_ticket_count(self):
        if self.ticket_collection is None:
            logger.error('Ticket Collection Does Not Exist')
            return 0
        username = self.data['username']
        result = self.ticket_collection.aggregate([
            {'$match': {'username': username}},
            {'$project': {self.current_month: 1, '_id': 0}}
        ])
        result_list = list(result)
        ticket_count = result_list[0].get(self.current_month, 0) if len(result_list) > 0 else 0
        logger.info(f'Ticket Count for {username} {ticket_count}')
        return ticket_count

    def is_paying_user(self):
        if self.ticket_collection is None:
            logger.error('Ticket Collection Does Not Exist')
            return False
        username = self.data['username']
        result = self.ticket_collection.find_one({'username': username})
        return result.get('is_paying_user', False) if result else False

    def use_faster_model(self):
        if self.ticket_collection is None:
            logger.error('Ticket Collection Does Not Exist')
            return True
        if self.is_paying_user():
            return self.get_ticket_count() >= 60
        return self.get_ticket_count() >= 3


def discord_log_error(content):
    try:
        url = DISCORD_WEBHOOK_URL
        data = {'content': content}
        headers = {'Content-Type': 'application/json'}
        response = requests.post(url, data=json.dumps(data), headers=headers)
        # Success: response.status_code == 204:
    except Exception as e:
        logger.error(f'Could not log to Discord: {e}')
        pass
